# encoding: utf-8
"""
@author: red0orange
@file: file.py
@desc: 文件相关的一些通用utils
"""
import os
import mimetypes
from pathlib import Path


def ifnone(a,b):
    "`a` if `a` is not None, otherwise `b`."
    return b if a is None else a

def poxis2str(l):
    return [str(i) for i in l]

def _path_to_same_str(p_fn):
    "path -> str, but same on nt+posix, for alpha-sort only"
    s_fn = str(p_fn)
    s_fn = s_fn.replace('\\','.')
    s_fn = s_fn.replace('/','.')
    return s_fn


def _get_files(parent, p, f, extensions):
    p = Path(p)#.relative_to(parent)
    if isinstance(extensions,str): extensions = [extensions]
    low_extensions = [e.lower() for e in extensions] if extensions is not None else None
    res = [p/o for o in f if not o.startswith('.')
           and (extensions is None or f'.{o.split(".")[-1].lower()}' in low_extensions)]
    return res


def get_files(path, extensions=None, recurse:bool=False, exclude=None,
              include=None, presort:bool=False, followlinks:bool=False):
    "Return list of files in `path` that have a suffix in `extensions`; optionally `recurse`."
    if recurse:
        res = []
        for i,(p,d,f) in enumerate(os.walk(path, followlinks=followlinks)):
            # skip hidden dirs
            if include is not None and i==0:   d[:] = [o for o in d if o in include]
            elif exclude is not None and i==0: d[:] = [o for o in d if o not in exclude]
            else:                              d[:] = [o for o in d if not o.startswith('.')]
            res += _get_files(path, p, f, extensions)
        if presort: res = sorted(res, key=lambda p: _path_to_same_str(p), reverse=False)
        return poxis2str(res)
    else:
        f = [o.name for o in os.scandir(path) if o.is_file()]
        res = _get_files(path, path, f, extensions)
        if presort: res = sorted(res, key=lambda p: _path_to_same_str(p), reverse=False)
        return poxis2str(res)


def get_image_files(path, extensions=None, recurse:bool=False, exclude=None,include=None, presort:bool=False, followlinks:bool=False):
    image_extensions = set(k for k, v in mimetypes.types_map.items() if v.startswith('image/'))
    extensions = ifnone(extensions, image_extensions)
    return get_files(path, extensions, recurse, exclude, include, presort, followlinks)


def get_image_extensions():
    return list(set(k for k, v in mimetypes.types_map.items() if v.startswith('image/')))



# encoding: utf-8
"""
@author: red0orange
@file: file.py
@desc: 其他一些utils
"""
from terminaltables import AsciiTable
import csv


def to_format_str(num):
    return '{:^7.3f}'.format(num)


def get_table_str(table_data):
    return AsciiTable(table_data).table


def read_csv(csv_path):
    "读出csv变成一个二维列表"
    result = []
    with open(csv_path, 'r') as f:
        reader = csv.reader(f)
        for row in reader:
            if len(row) != 0:
                result.append(row)
    return result


class BaseMeter(object):
    def __init__(self, name):
        self.name = name
        pass

    def return_data(self):
        raise BaseException('must override this func')


class AverageMeter(BaseMeter):
    """Computes and stores the average and current value"""
    def __init__(self, name):
        super(AverageMeter, self).__init__(name)
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def return_data(self):
        return self.avg


class AccMeter(BaseMeter):
    def __init__(self, name):
        super(AccMeter, self).__init__(name=name)
        self.predict_all_count = -1
        self.predict_true_count = -1
        self.acc = -1
        pass

    def reset(self):
        self.predict_all_count = -1
        self.predict_true_count = -1
        self.acc = -1
        pass

    def update(self, true_count, all_count):
        self.predict_all_count += all_count
        self.predict_true_count += true_count
        self.acc = self.predict_true_count / self.predict_all_count
        pass

    def return_data(self):
        return self.acc


# 数学方面
def get_list_intersection(listA, listB):
    return [i for i in listA if i in listB]


def get_list_subtraction(listA, listB):
    """
    求差集，得到listB中有而listA中没有的元素
    Args:
        listA:
        listB:

    Returns:

    """
    return list(set(listB).difference(set(listA)))


def get_list_union(listA, listB):
    return list(set(listA).union(set(listB)))


def if_list_contain_list(listA, listB):
    """
    判断listB是否包含listA
    Args:
        listA:
        listB:

    Returns:

    """
    return set(get_list_union(listA, listB)) == set(listB)



#!/home/hdh3/anaconda3/bin/python
# encoding: utf-8
"""
@author: red0orange
@file: args.py
@time:  11:10 PM
@desc:
"""
import argparse
import ast
import sys
import os
import time
import numpy as np
import torch
import shutil
import random


class SingleModel(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '__instance__'):
            cls.__instance__ = super(SingleModel, cls).__new__(cls)
        return cls.__instance__

    def init(self):
        pass


def get_class_attr(c):
    name_list = [i for i in dir(c) if not i.startswith('__')]
    result = {k: getattr(c, k) for k in name_list}
    return result


class BaseOption(object):
    class Arg:
        custom_data = True
        pass

    class Help:
        outcome_root = "实验结果保存的根目录"
        gpu = "选择使用的gpu"
        seed = "设置使用的随机种子"

    class ExtraArg:
        outcome_root = 'outcomes'
        gpu = "0"
        seed = 28
        description = ''
        name = ''


class Args(SingleModel):
    def init(self, option, expand_mode=False, code_mode=False):
        self.expand_mode = expand_mode
        self.code_mode = code_mode
        self.parse_args(option)

        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed_all(self.seed)
        random.seed(self.seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.enabled = True

        os.environ["CUDA_VISIBLE_DEVICES"] = self.gpu

        if self.name == '':
            self.model = '' if not hasattr(self, 'model') else self.self.model
            self.name = self.model + time.strftime('(%m-%d_%H:%M)', time.localtime(time.time()))
        else:
            self.name += time.strftime('(%m-%d_%H:%M)', time.localtime(time.time()))

        # 根据参数中的outcome_root、description、name的得到结果的保存文件夹路径
        if self.description == '':
            self.save_root = os.path.join(self.outcome_root, self.name)
        else:
            self.save_root = os.path.join(self.outcome_root, self.description, self.name)
        # 若本身已存在这个路径，就删除（一般情况下不会删除，因为有加时间戳），然后再创建这个文件夹
        if os.path.exists(self.save_root):
            shutil.rmtree(self.save_root)
        os.makedirs(self.save_root)

        return self

    def parse_args(self, option):
        # 跑自己代码的两种模式
        # 模式一：代码模式，一般直接在IDE中运行，不想要外部设置参数
        # 模式二：外部输入模式，一般在终端运行代码，需要外部输入设置参数
        # 跑开源代码的模式
        # 一般开源代码都是封装成parser，然后一堆需要外部输入的参数，这时候外部输入是正常工作的，不用修改，主要是
        # 代码模式要利用这个工具回复正常，即可以直接很方便地在代码里面设置参数而不需要去更改它parser里面的默认值
        # ，这个模式定义为拓展模式，在拓展模式下就是想要在开源代码中使用代码模式，直接在代码中设置参数
        expand_mode = self.expand_mode  # 设置为拓展模式，模拟外部输入参数运行开源代码
        code_mode   = self.code_mode    # 设置为代码模式，不从外部读取参数

        class_attr = get_class_attr(option)
        parse_args = get_class_attr(class_attr["Arg"])
        parse_help = get_class_attr(class_attr["Help"])
        parse_extra_args = get_class_attr(class_attr["ExtraArg"])

        parse_extra_args_names = list(parse_extra_args.keys())
        assert "outcome_root" in parse_extra_args_names, "必须设置实验结果保存的根目录"

        if expand_mode:
            for arg_name, arg_value in parse_args.items():
                sys.argv.extend(["--{}".format(arg_name), "{}".format(arg_value)])
            args_ = parse_args
        else:
            if not code_mode:
                parser = argparse.ArgumentParser()
                for arg_name in parse_args.keys():
                    arg_type = type(parse_args[arg_name])
                    arg_default_value = parse_args[arg_name]
                    arg_help = "" if parse_help.get(arg_name) is None else str(parse_help.get(arg_name))
                    if arg_type == bool:
                        parser.add_argument(f'--{arg_name}', type=ast.literal_eval, default=arg_default_value,
                                            help=arg_help)
                    else:
                        parser.add_argument(f'--{arg_name}', type=arg_type, default=arg_default_value, help=arg_help)
                args_ = vars(parser.parse_args())
            else:
                args_ = parse_args

        inter_args = [i for i in list(args_.keys()) if i in list(parse_extra_args.keys())]
        assert len(inter_args) == 0, "普通参数与额外参数不能重复"

        for arg_name, arg_value in {**args_, **parse_extra_args}.items():
            setattr(self, arg_name, arg_value)
            pass

    def get_args(self):
        return {k: getattr(self, k) for k in self.__dict__.keys() if not k.startswith(('_', '__'))}


option = Args()

if __name__ == '__main__':
    option.init(BaseOption, expand_mode=True, code_mode=True)

    print('==========================================')
    parser = argparse.ArgumentParser()
    parser.add_argument(f'--custom_data', type=str, default="0, 1", help="")
    args_ = vars(parser.parse_args())
    for key, value in args_.items():
        print(key, value)
    print('==========================================')
    for value in sys.argv:
        print(value)
    print('==========================================')
    for key, value in get_class_attr(option).items():
        print(key, value)
    pass



#!/home/hdh3/anaconda3/bin/python
# encoding: utf-8
"""
@author: red0orange
@file: database.py
@time:  7:35 PM
@desc:
"""
import datetime
import pymysql
from red0orange.utils import *


class DatabaseOperator(object):
    """
    封装对数据库的基本操作
    """
    def __init__(self, host_ip, user_name, user_pass, db_name):
        try:
            self._db = pymysql.connect(host_ip, user_name, user_pass, db_name)
        except:
            raise BaseException("连接失败，检查参数")

        self.cursor = self.db.cursor()
        pass

    @property
    def db(self):
        self._db.ping(reconnect=True)
        return self._db

    @classmethod
    def create_by_default(cls):
        """
        默认参数的db，这个是完全由代码指定
        Returns:

        """
        return cls("47.97.157.150", "red0orange", "asdfg123", "blood")

    @classmethod
    def create_by_normal(cls, host_ip, user_name, user_pass, db_name):
        """
        普通构建对象的方法
        Args:
            host_ip:
            user_name:
            user_pass:
            db_name:

        Returns:

        """
        return cls(host_ip, user_name, user_pass, db_name)

    @staticmethod
    def _get_column_names_types(cursor, table_name):
        sql = 'select column_name,data_type from information_schema.columns where table_name = "{}"'.format(table_name)
        cursor.execute(sql)
        names_types = cursor.fetchall()
        return [i[0] for i in names_types], [i[1] for i in names_types]

    @staticmethod
    def _get_column_names(cursor, table_name):
        return DatabaseOperator._get_column_names_types(cursor, table_name)[0]

    @staticmethod
    def _get_column_types(cursor, table_name):
        return DatabaseOperator._get_column_names_types(cursor, table_name)[1]

    def get_column_names(self, table_name):
        return self._get_column_types(self.cursor, table_name)

    def get_column_types(self, table_name):
        return self._get_column_types(self.cursor, table_name)

    @staticmethod
    def _get_column_type(cursor, table_name, column_name):
        """
        获取某个table的某一列的数据类型
        Args:
            cursor:
            table_name:
            column_name:

        Returns:

        """
        sql = f'select data_type from information_schema.columns where table_name="{table_name}" and column_name="{column_name}"'
        cursor.execute(sql)
        return cursor.fetchall()[0][0]

    @staticmethod
    def _modify_column_position(cursor, table_name, column_name, front_name=None):
        """
        改变数据表内某一列的位置
        Args:
            cursor:
            table_name:
            column_name:
            front_name:

        Returns:

        """
        column_type = DatabaseOperator._get_column_type(cursor, table_name, column_name)
        # TODO 这个问题要解决，现在是手动给varchar加东西
        # if column_type == 'varchar':
        #     column_type += '(10)'
        if front_name is None:
            sql = f'alter table {table_name} modify {column_name} {column_type} first'
            cursor.execute(sql)
        else:
            sql = f'alter table {table_name} modify {column_name} {column_type} after {front_name}'
            cursor.execute(sql)
        pass

    @staticmethod
    def _push_column_back(cursor, table_name, column_name, column_type):
        """
        在一个表后面加新字段
        Args:
            cursor:
            table_name:
            column_name:
            column_type:

        Returns:

        """
        sql = 'alter table {} add column {} {}'.format(table_name, column_name, column_type)
        cursor.execute(sql)
        pass

    @staticmethod
    def _modify_column_order(cursor, table_name, column_list):
        """
        按照column_list给table的column更换位置
        Args:
            cursor:
            table_name:
            column_list:

        Returns:

        """
        assert len(column_list) > 0, 'column list can not be empty'
        cur_list = DatabaseOperator._get_column_names(cursor, table_name)
        diff = get_list_subtraction(cur_list, column_list)
        for column_name in diff:
            # TODO 目前默认全部为varchar也就是不定长字符串类型
            DatabaseOperator._push_column_back(cursor, table_name, column_name, "text")

        last_name = None
        for i, column_name in enumerate(column_list):
            DatabaseOperator._modify_column_position(cursor, table_name, column_name, last_name)
            last_name = column_name

        pass

    @staticmethod
    def _insert_raw(cursor, table_name, data_dict):
        keys = list(data_dict.keys())
        values = list(data_dict.values())
        data_field_str = ','.join(keys)
        data_value_str = ','.join(['"' + str(i) + '"' for i in values])
        sql = 'INSERT INTO {} ({}) VALUES ({})'.format(table_name, data_field_str, data_value_str)
        cursor.execute(sql)
        pass

    def modify_column_order(self, table_name, column_list):
        self._modify_column_order(self.cursor, table_name, column_list)
        self.db.commit()
        pass

    def insert_raw(self, table_name, data_dict):
        self._insert_raw(self.cursor, table_name, data_dict)
        self.db.commit()
        pass


class ExpResult(object):
    def __init__(self):
        self._init = False

        self.column_list = None
        self.db = None
        self.cur_dict = None
        self.dt = None
        self.table_name = None
        pass

    def init(self, table_name, column_list, description, id):
        self._init = True
        self.table_name = table_name

        self.db = DatabaseOperator.create_by_default()
        self.column_list = column_list
        self.column_list.insert(0, 'datetime')
        self.column_list.insert(1, 'description')
        self.column_list.insert(2, 'id')
        self.db.modify_column_order(table_name, self.column_list)

        self.cur_dict = {k: None for k in column_list}

        self.dt = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.db.insert_raw(table_name, {'datetime': self.dt, 'description': description, 'id': id})
        pass

    def update(self, update_dict):
        if self._init:
            keys = list(update_dict.keys())
            assert if_list_contain_list(keys, self.column_list), "传入的更新dict，必须属于初始化时的列表"
            self.cur_dict.update(update_dict)

            for key in keys:
                # TODO 目前直接固定按照datetime索引进行更新，应该有更好的方式
                sql = 'update {} set {}="{}" where datetime="{}"'.format(self.table_name, key, update_dict[key], self.dt)
                self.db.cursor.execute(sql)
            self.db.db.commit()

            pass
        else:
            raise BaseException('not init')


exp_result = ExpResult()



#!/home/hdh3/anaconda3/bin/python
# encoding: utf-8
"""
@author: red0orange
@file: recorder.py
@time:  7:32 PM
@desc:
"""
import traceback
from runx.logx import LogX
import shutil
import os
import torch


class Recorder(LogX):
    def __init__(self):
        super(Recorder, self).__init__()
        self.record_dir = None
        self.save_last = None

        self._init = False
        pass

    def init(self, save_root, save_modules=[]):
        # init recorder
        self.initialize(record_dir=save_root, save_last=5, modules_need_to_save=save_modules,
                        coolname=True, tensorboard=True, eager_flush=True)
        self._init = True
        pass

    def initialize(self, record_dir=None, save_last=5, modules_need_to_save=[], coolname=False, hparams=None,
                   tensorboard=False, no_timestamp=False, global_rank=0,
                   eager_flush=True):
        # 调用runx中的初始化，保存路径为record_dir/metrics
        super(Recorder, self).initialize(logdir=os.path.join(record_dir, 'metrics'), coolname=coolname, hparams=hparams,
                                         tensorboard=tensorboard, no_timestamp=no_timestamp, global_rank=global_rank,
                                         eager_flush=eager_flush)
        self.record_dir = record_dir
        self.save_last = save_last

        # 创建需要的子文件夹
        os.makedirs(os.path.join(self.record_dir, 'models'), exist_ok=True)
        os.makedirs(os.path.join(self.record_dir, 'metrics'), exist_ok=True)
        os.makedirs(os.path.join(self.record_dir, 'code'), exist_ok=True)

        # 保存传入的参数中需要保存的模块文件
        file_save_root = os.path.join(self.record_dir, 'code')
        file_paths = []
        for module in modules_need_to_save:
            try:
                file_path = module.__file__
            except Exception:
                raise BaseException('error load module file to save')
            file_paths.append(file_path)
        # 调用这个函数的主文件默认保存
        file_paths.append(traceback.extract_stack()[0].filename)
        file_paths = [os.path.abspath(i) for i in file_paths]
        file_paths = list(set(file_paths))
        for file_path in file_paths:
            new_path = os.path.join(file_save_root, os.path.basename(file_path))
            shutil.copy(file_path, new_path)
        pass

    def save_model(self, save_dict, epoch, best=False):
        if not self._init:
            raise BaseException("not init")
        # 使用自己的保存方法
        save_root = os.path.join(self.record_dir, 'models')
        save_path = os.path.join(save_root, f'{epoch}.pth')
        torch.save(save_dict, save_path)
        if not best:
            if os.path.exists(os.path.join(save_root, f'{epoch-self.save_last}.pth')):
                os.remove(os.path.join(save_root, f'{epoch-self.save_last}.pth'))
        pass


recorder = Recorder()



# encoding: utf-8
"""
@author: red0orange
@file: file.py
@desc: 绘图相关的一些通用utils
"""
import torch
import os
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns


def create_fig(figsize=(8,8), row=1, col=1):
    fig = plt.figure(figsize=figsize)
    fig.tight_layout()
    axes = fig.subplots(row, col)
    return fig, axes


def plot_bars(axe, x_labels, ys, y_labels=None, bar_width=0.28):
    if not isinstance(ys[0], list):
        ys = [ys]
    if y_labels is None:
        y_labels = [str(i) for i in range(len(ys))]
    if not isinstance(y_labels, list):
        y_labels = [y_labels]
    x_labels = [str(i) for i in x_labels]
    
    bar_num = len(ys)

    x = np.arange(len(x_labels))  # the label locations
    width = bar_num * bar_width  # the width of the bars

    rects = []
    for i in range(bar_num):
        rects.append(axe.bar(x - width/2 + bar_width*(i+0.5), ys[i], width/bar_num, label=y_labels[i]))

    # Add some text for labels, title and custom x-axis tick labels, etc.
    axe.set_ylabel('Scores')
    axe.set_xticks(x)
    axe.set_xticklabels(x_labels)
    axe.legend()

    def autolabel(rects):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects:
            height = rect.get_height()
            axe.annotate('{}'.format(height),
                        xy=(rect.get_x() + rect.get_width() / 2, height),
                        xytext=(0, 3),  # 3 points vertical offset
                        textcoords="offset points",
                        ha='center', va='bottom')

    for rect in rects:
        autolabel(rect)
        



# encoding: utf-8
"""
@author: red0orange
@file: preprocess.py
@desc: 数据预处理可能会用到的utils
"""
import os
import csv
import numpy as np
import xml.dom.minidom as minidom
from collections import Counter, OrderedDict
import warnings


def modify_xml_to_correct(root_path, new_root, class_dict):
    "临时函数更改所有xml文件用于适应代码的变化"
    paths = get_files(root_path, ['.xml'], recurse=True)
    for path in paths:
        try:
            rel_path = os.path.relpath(path, root_path)
            old_path = path
            new_path = os.path.join(new_root, rel_path)
            print(old_path)
            if not os.path.isdir(os.path.dirname(new_path)):
                os.makedirs(os.path.dirname(new_path), exist_ok=True)
            with open(old_path, 'r', encoding='utf-8') as f:
                dom = minidom.parse(f)
                root = dom.documentElement

                filename = root.getElementsByTagName('filename')[0].firstChild.data
                fileLocation_ = root.getElementsByTagName('fileLocation')[0].firstChild.data
                relativeLocation = root.getElementsByTagName('relativeLocation')[0].firstChild.data
                img_width_ = root.getElementsByTagName('width')[0].firstChild.data
                img_height_ = root.getElementsByTagName('height')[0].firstChild.data

                object = root.getElementsByTagName('object')
                if len(object) != 1:
                    raise BaseException('xml error')
                object = object[0]
                names = object.getElementsByTagName('name')
                names = [name_node.firstChild.data for name_node in names]
                xs = object.getElementsByTagName('x')
                xs = [x_node.firstChild.data for x_node in xs]
                ys = object.getElementsByTagName('y')
                ys = [y_node.firstChild.data for y_node in ys]
                widths = object.getElementsByTagName('width')
                widths = [width_node.firstChild.data for width_node in widths]
                heights = object.getElementsByTagName('height')
                heights = [hegiht_node.firstChild.data for hegiht_node in heights]

            # 重新写
            doc = minidom.Document()
            annotation = doc.createElement('annotation')
            doc.appendChild(annotation)

            file_path = doc.createElement('file_path')
            file_path.appendChild(doc.createTextNode(fileLocation_))
            annotation.appendChild(file_path)

            OneDirname = doc.createElement('patientNum')  # 上一级的目录名
            OneDirname.appendChild(doc.createTextNode(os.path.dirname(fileLocation_).split('/')[-1]))
            TwoDirname = doc.createElement('illSort')  # 上二级的目录名
            TwoDirname.appendChild(doc.createTextNode(os.path.dirname(fileLocation_).split('/')[-2]))
            annotation.appendChild(OneDirname)
            annotation.appendChild(TwoDirname)

            size = doc.createElement('img_size')
            annotation.appendChild(size)
            width = doc.createElement('width')
            width.appendChild(doc.createTextNode('{}'.format(img_width_)))
            height = doc.createElement('height')
            height.appendChild(doc.createTextNode('{}'.format(img_height_)))
            size.appendChild(width)
            size.appendChild(height)

            for x_, y_, width_, height_, name_ in zip(xs, ys, widths, heights, names):
                _object = doc.createElement('object')
                annotation.appendChild(_object)

                # 已经错误了的名字
                name_ = name_.replace('淋巴细胞', '幼淋巴细胞').replace('嗜酸晚幼粒细胞', '嗜酸性粒细胞')\
                    .replace('幼幼淋巴细胞', '幼淋巴细胞').replace('异型幼淋巴细胞', '幼淋巴细胞').replace('原幼稚幼淋巴细胞', '原幼稚淋巴细胞')\
                    .replace('成熟幼淋巴细胞', '成熟淋巴细胞').replace('晩幼粒细胞', '晚幼粒细胞')

                x_, y_, width_, height_, img_width_, img_height_ = map(int, [x_, y_, width_, height_, img_width_,
                                                                             img_height_])
                x_, width_ = map(lambda i: '{:.3}'.format(i / img_width_), [x_, width_])
                y_, height_ = map(lambda i: '{:.3}'.format(i / img_height_), [y_, height_])
                categories_id = doc.createElement('categories_id')
                categories_id.appendChild(doc.createTextNode(str(class_dict[name_])))
                _object.appendChild(categories_id)
                name = doc.createElement('name')
                name.appendChild(doc.createTextNode(name_))
                _object.appendChild(name)
                x = doc.createElement('x')
                x.appendChild(doc.createTextNode(x_))
                y = doc.createElement('y')
                y.appendChild(doc.createTextNode(y_))
                width = doc.createElement('width')
                width.appendChild(doc.createTextNode(width_))
                height = doc.createElement('height')
                height.appendChild(doc.createTextNode(height_))
                _object.appendChild(x)
                _object.appendChild(y)
                _object.appendChild(width)
                _object.appendChild(height)

            with open(new_path, 'w', encoding='utf-8') as f:
                doc.writexml(f, indent='', addindent='\t', newl='\n', encoding='utf-8')
        except:
            print(old_path)

            
def parse_xml_to_boxes(xml_path):
    boxes = []
    names = []
    with open(xml_path, 'r', encoding='utf-8') as f:
        dom = minidom.parse(f)
        root = dom.documentElement
        objects = root.getElementsByTagName('object')
        for object_ in objects:
            categories_id = object_.getElementsByTagName('categories_id')[0].firstChild.data
            name = object_.getElementsByTagName('name')[0].firstChild.data
            x = object_.getElementsByTagName('x')[0].firstChild.data
            y = object_.getElementsByTagName('y')[0].firstChild.data
            width = object_.getElementsByTagName('width')[0].firstChild.data
            height = object_.getElementsByTagName('height')[0].firstChild.data
            categories_id = int(categories_id)
            x, y, width, height = map(float, [x, y, width, height])
            boxes.append([int(categories_id), x, y, width, height])
            names.append(name)
    return np.array(boxes), names
            

def change_xml(old_root, new_root, xml_paths, class_transform=None):
    "转换新的xml格式的函数"
    for xml_path in xml_paths:
        rel_path = os.path.relpath(xml_path, old_root)
        old_path = xml_path
        new_path = os.path.join(new_root, rel_path)
#         print(new_path)
        new_path = new_path.rsplit('.', maxsplit=1)[0]+'.txt'
        if not os.path.isdir(os.path.dirname(new_path)):
            os.makedirs(os.path.dirname(new_path), exist_ok=True)
        boxes, names = parse_xml_to_boxes(old_path)

        # 转换格式
        if len(boxes.shape) == 1:
            print('empty label: ', old_path)
            boxes = boxes[None, ...]
            with open(new_path, 'w') as txt:
                pass
            continue
        boxes[:, 1] = boxes[:, 1] + boxes[:, 3] / 2
        boxes[:, 2] = boxes[:, 2] + boxes[:, 4] / 2
#         print(new_path)
        with open(new_path, 'w') as txt:
            for box_i in range(boxes.shape[0]):
                cat_id, x, y, width, height = boxes[box_i, :]
                if class_transform:
                    cat_id = class_transform[cat_id]
                txt.write('{} {:.3} {:.3} {:.3} {:.3}'.format(cat_id, x, y, width, height) + '\n')


def get_valid_image_label(image_root, lebel_root):
    "从图片的根目录和label的根目录获得无重名的image->label的文件对应字典"
    image_extensions = set(k for k, v in mimetypes.types_map.items() if v.startswith('image/'))
    img_paths = get_files(image_root, extensions=image_extensions, recurse=True)
    label_paths = get_files(lebel_root, extensions=['.txt'], recurse=True)

    img_names = [os.path.basename(path).split('.')[0] for path in img_paths]
    label_names = [os.path.basename(path).split('.')[0] for path in label_paths]

    img_label = {}
    for i, img_name in enumerate(img_names):
        img_label[img_paths[i]] = []
        for ii, label_name in enumerate(label_names):
            if img_name == label_name:
                img_label[img_paths[i]].append(label_paths[ii])

    valid_img_label = {k: v[0] for k, v in img_label.items() if len(v) == 1}
    return valid_img_label


def get_samples_statistics(image_paths, label_paths):
    "传入对应的图片路径列表和标注路径列表"
    image_info_result = {}
    class_info_result = {}
    for i, (image_path, label_path) in enumerate(zip(image_paths, label_paths)):
        boxes = np.loadtxt(label_path)
        if len(boxes) == 0:
            image_info_result[image_path] = {}
            continue
        if len(boxes.shape) == 1: boxes = boxes[None, ...]
        labels = boxes[:, 0].astype(np.int64)
        cls_counter = dict(Counter(labels))
        image_info_result[image_path] = cls_counter
        for key, value in cls_counter.items():
            if key not in class_info_result.keys():
                class_info_result[key] = {}
            class_info_result[key][image_path] = value
    return image_info_result, class_info_result


def per_class_split(class_info_result, image_info_result, test_ratio=0.15, valid_ratio=0.1):
    train_image_paths = []
    test_image_paths = []
    valid_image_paths = []

    sorted_class_info = [[cls_id, info] for cls_id, info in class_info_result.items()]
    sorted_class_info = sorted(sorted_class_info, key=lambda k: sum(list(k[1].values())))

    used_image_paths = []

    # 得到每个类别每个数据集需要的数量
    cls_split = [{}, {}, {}, {}]  # all   train  test  valid
    for cls_id, info in sorted_class_info:
        # sort the dict
        cls_num = 0
        sorted_info = []
        for image_path, num in info.items():
            sorted_info.append([image_path, num])
            cls_num += num
        sorted_info = sorted(sorted_info, key=lambda k: k[1])

        train_ratio = 1 - test_ratio - valid_ratio
        train_num = round(cls_num * train_ratio)
        test_num = round((cls_num - train_num) * (test_ratio / (test_ratio + valid_ratio)))
        valid_num = round((cls_num - train_num) * (valid_ratio / (test_ratio + valid_ratio)))

        cls_split[0][cls_id] = cls_num
        cls_split[1][cls_id] = train_num
        cls_split[2][cls_id] = test_num
        cls_split[3][cls_id] = valid_num

    # 不断更新的实际分布情况
    res_cls_split = [cls_split[0],
                     {k: 0 for k in cls_split[1].keys()},
                     {k: 0 for k in cls_split[2].keys()},
                     {k: 0 for k in cls_split[3].keys()}]
    # 投票分每张图片
    for i, (image_path, info) in enumerate(image_info_result.items()):
        # 得到这张图片的类别情况
        exist_cls, exist_cls_num = info.keys(), info.values()
        if len(exist_cls) == 0:
            warnings.warn('empty image: {}'.format(image_path))
            continue
        # 计算谁最合适得到这张图片
        train_error_all, test_error_all, valid_error_all = [], [], []
        for cls_id in exist_cls:
            if cls_split[1][cls_id] != 0:
                train_error_all.append(
                    100 * abs((cls_split[1][cls_id] - res_cls_split[1][cls_id]) / cls_split[1][cls_id]))
            if cls_split[2][cls_id] != 0:
                test_error_all.append(
                    100 * abs((cls_split[2][cls_id] - res_cls_split[2][cls_id]) / cls_split[2][cls_id]))
            if cls_split[3][cls_id] != 0:
                valid_error_all.append(
                    100 * abs((cls_split[3][cls_id] - res_cls_split[3][cls_id]) / cls_split[3][cls_id]))
        train_mean, test_mean, valid_mean = map(lambda k: sum(k)/len(k) if len(k) != 0 else 0,
                                                [train_error_all, test_error_all, valid_error_all])
        if max([train_mean, test_mean, valid_mean]) == train_mean:
            suitable_dataset = train_image_paths
            suitable_cls_split = res_cls_split[1]
        elif max([train_mean, test_mean, valid_mean]) == test_mean:
            suitable_dataset = test_image_paths
            suitable_cls_split = res_cls_split[2]
        elif max([train_mean, test_mean, valid_mean]) == valid_mean:
            suitable_dataset = valid_image_paths
            suitable_cls_split = res_cls_split[3]
        suitable_dataset.append(image_path)
        for cls_id, cls_num in zip(exist_cls, exist_cls_num):
            suitable_cls_split[cls_id] += cls_num
    return train_image_paths, valid_image_paths, test_image_paths, res_cls_split, cls_split


def read_class_dict_csv(path):
    f = open(path, 'r')
    reader = csv.reader(f)
    result = {}
    for line in reader:
        if len(line) == 0:
            continue
        result[int(line[0])] = line[1]
    f.close()
    return result

    
def write_csv(image_label, csv_path):
    with open(csv_path, 'w') as f:
        writer = csv.writer(f)
        for key, value in image_label.items():
            writer.writerow([str(key), str(value)])
    pass

def filter_samples(image_paths, label_paths):
    img_xml = {}
    image_names = [os.path.basename(path).split('.')[0] for path in image_paths]
    label_names = [os.path.basename(path).split('.')[0] for path in label_paths]
    for i, image_name in enumerate(image_names):
        img_xml[image_paths[i]] = []
        for ii, lable_name in enumerate(label_names):
            if image_name == lable_name:
                img_xml[image_paths[i]].append(label_paths[ii])
    valid_img_xml = {k:v[0] for k, v in img_xml.items() if len(v)==1}
    unvalid_img_xml = {k:v for k, v in img_xml.items() if len(v)!=1}
    return valid_img_xml, unvalid_img_xml